@@ -8,7 +8,7 @@ use std::pin::Pin;
 use std::sync::{Arc, Mutex};
 
 use async_trait::async_trait;
-use futures::future::{self, Future};
+use futures::future::Future;
 use futures::task::{Context, Poll, Waker};
 use log::debug;
 
@@ -16,56 +16,13 @@ use tc_error::*;
 
 use super::{Transact, TxnId};
 
-/// Define a way to manage transaction-specific versions of a state.
-#[async_trait]
-pub trait Mutate: Send {
-    /// The type of this state when pending a transaction commit.
-    type Pending: Clone + Send;
-
-    /// Create a new transaction-specific version of this state.
-    fn diverge(&self, txn_id: &TxnId) -> Self::Pending;
-
-    /// Canonicalize a transaction-specific version.
-    async fn converge(&mut self, new_value: Self::Pending);
-}
-
-/// A generic impl of [`Mutate`], for convenience.
-pub struct Mutable<T: Clone + Send> {
-    value: T,
-}
-
-impl<T: Clone + Send> Mutable<T> {
-    pub fn value(&'_ self) -> &'_ T {
-        &self.value
-    }
-}
-
-#[async_trait]
-impl<T: Clone + Send> Mutate for Mutable<T> {
-    type Pending = T;
-
-    fn diverge(&self, _txn_id: &TxnId) -> Self::Pending {
-        self.value.clone()
-    }
-
-    async fn converge(&mut self, new_value: Self::Pending) {
-        self.value = new_value;
-    }
-}
-
-impl<T: Clone + Send> From<T> for Mutable<T> {
-    fn from(value: T) -> Mutable<T> {
-        Mutable { value }
-    }
-}
-
 /// An immutable read guard for a transactional state.
-pub struct TxnLockReadGuard<T: Mutate> {
+pub struct TxnLockReadGuard<T> {
     txn_id: TxnId,
     lock: TxnLock<T>,
 }
 
-impl<T: Mutate> TxnLockReadGuard<T> {
+impl<T> TxnLockReadGuard<T> {
     /// Upgrade this read lock to a write lock.
     pub fn upgrade(self) -> TxnLockWriteFuture<T> {
         TxnLockWriteFuture {
@@ -75,37 +32,36 @@ impl<T: Mutate> TxnLockReadGuard<T> {
     }
 }
 
-impl<T: Mutate> Deref for TxnLockReadGuard<T> {
-    type Target = <T as Mutate>::Pending;
+impl<T> Deref for TxnLockReadGuard<T> {
+    type Target = T;
 
-    fn deref(&self) -> &<T as Mutate>::Pending {
+    fn deref(&self) -> &T {
         unsafe {
             &*self
                 .lock
                 .inner
+                .state
                 .lock()
-                .unwrap()
-                .value_at
+                .expect("TxnLock read")
+                .at
                 .get(&self.txn_id)
-                .unwrap()
+                .expect("TxnLock read version")
                 .get()
         }
     }
 }
 
-impl<T: Mutate> Drop for TxnLockReadGuard<T> {
+impl<T> Drop for TxnLockReadGuard<T> {
     fn drop(&mut self) {
-        let lock = &mut self.lock.inner.lock().unwrap();
-        match lock.state.readers.get_mut(&self.txn_id) {
+        let mut state = self.lock.inner.state.lock().expect("TxnLockReadGuard drop");
+        match state.readers.get_mut(&self.txn_id) {
             Some(count) if *count > 1 => (*count) -= 1,
             Some(count) if *count == 1 => {
-                (*count) -= 1;
+                *count = 0;
 
-                while let Some(waker) = lock.state.wakers.pop_front() {
+                while let Some(waker) = state.wakers.pop_front() {
                     waker.wake()
                 }
-
-                lock.state.wakers.shrink_to_fit();
             }
             _ => panic!("TxnLockReadGuard count updated incorrectly!"),
         }
@@ -113,12 +69,12 @@ impl<T: Mutate> Drop for TxnLockReadGuard<T> {
 }
 
 /// An exclusive write lock for a transactional state.
-pub struct TxnLockWriteGuard<T: Mutate> {
+pub struct TxnLockWriteGuard<T> {
     txn_id: TxnId,
     lock: TxnLock<T>,
 }
 
-impl<T: Mutate> TxnLockWriteGuard<T> {
+impl<T: Clone> TxnLockWriteGuard<T> {
     /// Downgrade this write lock into a read lock;
     pub fn downgrade(self, txn_id: &'_ TxnId) -> TxnLockReadFuture<T> {
         if txn_id != &self.txn_id {
@@ -129,82 +85,88 @@ impl<T: Mutate> TxnLockWriteGuard<T> {
     }
 }
 
-impl<T: Mutate> Deref for TxnLockWriteGuard<T> {
-    type Target = <T as Mutate>::Pending;
+impl<T: Clone> Deref for TxnLockWriteGuard<T> {
+    type Target = T;
 
-    fn deref(&self) -> &<T as Mutate>::Pending {
+    fn deref(&self) -> &T {
         unsafe {
             &*self
                 .lock
                 .inner
+                .state
                 .lock()
-                .unwrap()
-                .value_at
+                .expect("TxnLock read")
+                .at
                 .get(&self.txn_id)
-                .unwrap()
+                .expect("TxnLock version read")
                 .get()
         }
     }
 }
 
-impl<T: Mutate> DerefMut for TxnLockWriteGuard<T> {
-    fn deref_mut(&mut self) -> &mut <T as Mutate>::Pending {
+impl<T: Clone> DerefMut for TxnLockWriteGuard<T> {
+    fn deref_mut(&mut self) -> &mut T {
         unsafe {
             &mut *self
                 .lock
                 .inner
+                .state
                 .lock()
-                .unwrap()
-                .value_at
+                .expect("TxnLock write")
+                .at
                 .get_mut(&self.txn_id)
-                .unwrap()
+                .expect("TxnLock version write")
                 .get()
         }
     }
 }
 
-impl<T: Mutate> Drop for TxnLockWriteGuard<T> {
+impl<T> Drop for TxnLockWriteGuard<T> {
     fn drop(&mut self) {
-        let lock = &mut self.lock.inner.lock().unwrap();
-        lock.state.reserved = None;
+        let mut state = self
+            .lock
+            .inner
+            .state
+            .lock()
+            .expect("TxnLockWriteGuard drop");
 
-        while let Some(waker) = lock.state.wakers.pop_front() {
+        state.reserved = None;
+
+        while let Some(waker) = state.wakers.pop_front() {
             waker.wake()
         }
-
-        lock.state.wakers.shrink_to_fit();
     }
 }
 
-struct LockState {
+struct LockState<T> {
     last_commit: Option<TxnId>,
     readers: BTreeMap<TxnId, usize>,
     reserved: Option<TxnId>,
     wakers: VecDeque<Waker>,
+
+    canon: UnsafeCell<T>,
+    at: BTreeMap<TxnId, UnsafeCell<T>>,
 }
 
-struct Inner<T: Mutate> {
-    state: LockState,
-    value: UnsafeCell<T>,
-    value_at: BTreeMap<TxnId, UnsafeCell<<T as Mutate>::Pending>>,
+struct Inner<T> {
+    name: String,
+    state: Mutex<LockState<T>>,
 }
 
 /// A lock which provides transaction-specific versions of the locked state.
-pub struct TxnLock<T: Mutate> {
-    name: String,
-    inner: Arc<Mutex<Inner<T>>>,
+pub struct TxnLock<T> {
+    inner: Arc<Inner<T>>,
 }
 
-impl<T: Mutate> Clone for TxnLock<T> {
+impl<T> Clone for TxnLock<T> {
     fn clone(&self) -> Self {
         TxnLock {
-            name: self.name.clone(),
             inner: self.inner.clone(),
         }
     }
 }
 
-impl<T: Mutate> TxnLock<T> {
+impl<T: Clone> TxnLock<T> {
     /// Create a new lock.
     pub fn new<I: fmt::Display>(name: I, value: T) -> TxnLock<T> {
         let state = LockState {
@@ -212,55 +174,57 @@ impl<T: Mutate> TxnLock<T> {
             readers: BTreeMap::new(),
             reserved: None,
             wakers: VecDeque::new(),
+
+            canon: UnsafeCell::new(value),
+            at: BTreeMap::new(),
         };
 
         let inner = Inner {
-            state,
-            value: UnsafeCell::new(value),
-            value_at: BTreeMap::new(),
+            name: name.to_string(),
+            state: Mutex::new(state),
         };
 
         TxnLock {
-            name: name.to_string(),
-            inner: Arc::new(Mutex::new(inner)),
+            inner: Arc::new(inner),
         }
     }
 
     /// Try to acquire a read lock synchronously.
     pub fn try_read(&self, txn_id: &TxnId) -> TCResult<Option<TxnLockReadGuard<T>>> {
-        let lock = &mut self.inner.lock().unwrap();
+        let mut state = if let Ok(state) = self.inner.state.try_lock() {
+            state
+        } else {
+            return Ok(None);
+        };
 
-        let last_commit = lock
-            .state
-            .last_commit
-            .as_ref()
-            .unwrap_or(&super::id::MIN_ID);
-        if !lock.value_at.contains_key(txn_id) && txn_id < last_commit {
+        let last_commit = state.last_commit.as_ref().unwrap_or(&super::id::MIN_ID);
+
+        if !state.at.contains_key(txn_id) && txn_id < last_commit {
             // If the requested time is too old, just return an error.
             // We can't keep track of every historical version here.
             debug!(
                 "transaction {} is already finalized, can't acquire read lock",
                 txn_id
             );
+
             Err(TCError::conflict())
-        } else if lock.state.reserved.is_some() && txn_id >= lock.state.reserved.as_ref().unwrap() {
+        } else if let Some(ref past_write) = state.reserved {
             // If a writer can mutate the locked value at the requested time, wait it out.
-            let past_write = lock.state.reserved.as_ref().unwrap();
             debug!(
                 "TxnLock {} is already reserved for writing at {}",
-                &self.name, past_write
+                &self.inner.name, past_write
             );
+
             Ok(None)
         } else {
             // Otherwise, return a ReadGuard.
-            if !lock.value_at.contains_key(txn_id) {
-                let value_at_txn_id =
-                    UnsafeCell::new(unsafe { (&*lock.value.get()).diverge(txn_id) });
+            if !state.at.contains_key(txn_id) {
+                let value_at_txn_id = UnsafeCell::new(unsafe { (&*state.canon.get()).clone() });
 
-                lock.value_at.insert(*txn_id, value_at_txn_id);
+                state.at.insert(*txn_id, value_at_txn_id);
             }
 
-            *lock.state.readers.entry(*txn_id).or_insert(0) += 1;
+            *state.readers.entry(*txn_id).or_insert(0) += 1;
 
             Ok(Some(TxnLockReadGuard {
                 txn_id: *txn_id,
@@ -279,33 +243,41 @@ impl<T: Mutate> TxnLock<T> {
 
     /// Try to acquire a write lock, synchronously.
     pub fn try_write(&self, txn_id: &TxnId) -> TCResult<Option<TxnLockWriteGuard<T>>> {
-        let lock = &mut self.inner.lock().unwrap();
-        let latest_reader = lock.state.readers.keys().max();
+        let mut state = if let Ok(state) = self.inner.state.try_lock() {
+            state
+        } else {
+            return Ok(None);
+        };
 
-        if latest_reader.is_some() && latest_reader.unwrap() > txn_id {
+        if let Some(latest_read) = state.readers.keys().max() {
             // If there's already a reader in the future, there's no point in waiting.
-            return Err(TCError::conflict());
+            if latest_read > txn_id {
+                return Err(TCError::conflict());
+            }
         }
 
-        match &lock.state.reserved {
+        match &state.reserved {
             // If there's already a writer in the future, there's no point in waiting.
             Some(current_txn) if current_txn > txn_id => Err(TCError::conflict()),
             // If there's a writer in the past, wait for it to complete.
             Some(current_txn) if current_txn < txn_id => {
                 debug!(
                     "TxnLock {} at {} blocked on {}",
-                    &self.name, txn_id, current_txn
+                    &self.inner.name, txn_id, current_txn
                 );
 
                 Ok(None)
             }
             // If there's already a writer for the current transaction, wait for it to complete.
             Some(_) => {
-                debug!("TxnLock {} waiting on existing write lock", &self.name);
+                debug!(
+                    "TxnLock {} waiting on existing write lock",
+                    &self.inner.name
+                );
                 Ok(None)
             }
             _ => {
-                if let Some(readers) = lock.state.readers.get(txn_id) {
+                if let Some(readers) = state.readers.get(txn_id) {
                     if readers > &0 {
                         // There's already a reader for the current transaction, wait for it.
                         return Ok(None);
@@ -313,10 +285,11 @@ impl<T: Mutate> TxnLock<T> {
                 }
 
                 // Otherwise, copy the value to be mutated in this transaction.
-                lock.state.reserved = Some(*txn_id);
-                if !lock.value_at.contains_key(txn_id) {
-                    let mutation = UnsafeCell::new(unsafe { (&*lock.value.get()).diverge(txn_id) });
-                    lock.value_at.insert(*txn_id, mutation);
+                state.reserved = Some(*txn_id);
+                if !state.at.contains_key(txn_id) {
+                    let mutation = UnsafeCell::new(unsafe { (&*state.canon.get()).clone() });
+
+                    state.at.insert(*txn_id, mutation);
                 }
 
                 Ok(Some(TxnLockWriteGuard {
@@ -337,69 +310,51 @@ impl<T: Mutate> TxnLock<T> {
 }
 
 #[async_trait]
-impl<T: Mutate> Transact for TxnLock<T> {
+impl<T: Clone + Send> Transact for TxnLock<T> {
     async fn commit(&self, txn_id: &TxnId) {
-        debug!("TxnLock::commit {} at {}", self.name, txn_id);
+        debug!("TxnLock::commit {} at {}", self.inner.name, txn_id);
 
-        async {
-            self.read(&txn_id).await.unwrap(); // make sure there's no active write lock
-            let lock = &mut self.inner.lock().unwrap();
-            assert!(lock.state.reserved.is_none());
-            if let Some(last_commit) = &lock.state.last_commit {
-                assert!(last_commit < &txn_id);
-            }
-
-            lock.state.last_commit = Some(*txn_id);
-
-            let new_value = if let Some(cell) = lock.value_at.get(&txn_id) {
-                let pending: &<T as Mutate>::Pending = unsafe { &*cell.get() };
-                Some(pending.clone())
-            } else {
-                None
-            };
-
-            #[allow(clippy::async_yields_async)]
-            if let Some(new_value) = new_value {
-                let value = unsafe { &mut *lock.value.get() };
-                value.converge(new_value)
-            } else {
-                Box::pin(future::ready(()))
-            }
+        let mut state = self.inner.state.lock().expect("TxnLock commit state");
+        assert!(state.reserved.is_none());
+        if let Some(ref last_commit) = state.last_commit {
+            assert!(last_commit < &txn_id);
         }
-        .await
-        .await;
 
-        let lock = &mut self.inner.lock().unwrap();
+        state.last_commit = Some(*txn_id);
 
-        while let Some(waker) = lock.state.wakers.pop_front() {
+        if let Some(cell) = state.at.get(txn_id) {
+            let value = unsafe { &mut *state.canon.get() };
+            let pending: &T = unsafe { &*cell.get() };
+            *value = pending.clone();
+        }
+
+        while let Some(waker) = state.wakers.pop_front() {
             waker.wake()
         }
-
-        lock.state.wakers.shrink_to_fit()
     }
 
     async fn finalize(&self, txn_id: &TxnId) {
-        debug!("TxnLock {} finalize {}", &self.name, txn_id);
+        debug!("TxnLock {} finalize {}", &self.inner.name, txn_id);
 
-        let lock = &mut self.inner.lock().unwrap();
+        let mut state = self.inner.state.lock().expect("TxnLock finalize state");
 
-        if let Some(readers) = lock.state.readers.remove(txn_id) {
+        if let Some(readers) = state.readers.remove(txn_id) {
             if readers > 0 {
                 panic!("tried to finalize a transaction that's still active!")
             }
         }
 
-        lock.value_at.remove(txn_id);
+        state.at.remove(txn_id);
     }
 }
 
 /// A read lock future.
-pub struct TxnLockReadFuture<'a, T: Mutate> {
+pub struct TxnLockReadFuture<'a, T> {
     txn_id: &'a TxnId,
     lock: TxnLock<T>,
 }
 
-impl<'a, T: Mutate> Future for TxnLockReadFuture<'a, T> {
+impl<'a, T: Clone> Future for TxnLockReadFuture<'a, T> {
     type Output = TCResult<TxnLockReadGuard<T>>;
 
     fn poll(self: Pin<&mut Self>, context: &mut Context) -> Poll<Self::Output> {
@@ -409,9 +364,9 @@ impl<'a, T: Mutate> Future for TxnLockReadFuture<'a, T> {
             Ok(None) => {
                 self.lock
                     .inner
-                    .lock()
-                    .unwrap()
                     .state
+                    .lock()
+                    .expect("TxnLockReadFuture state")
                     .wakers
                     .push_back(context.waker().clone());
 
@@ -422,12 +377,12 @@ impl<'a, T: Mutate> Future for TxnLockReadFuture<'a, T> {
 }
 
 /// A write lock future.
-pub struct TxnLockWriteFuture<T: Mutate> {
+pub struct TxnLockWriteFuture<T> {
     txn_id: TxnId,
     lock: TxnLock<T>,
 }
 
-impl<T: Mutate> Future for TxnLockWriteFuture<T> {
+impl<T: Clone> Future for TxnLockWriteFuture<T> {
     type Output = TCResult<TxnLockWriteGuard<T>>;
 
     fn poll(self: Pin<&mut Self>, context: &mut Context) -> Poll<Self::Output> {
@@ -437,9 +392,9 @@ impl<T: Mutate> Future for TxnLockWriteFuture<T> {
             Ok(None) => {
                 self.lock
                     .inner
-                    .lock()
-                    .unwrap()
                     .state
+                    .lock()
+                    .expect("TxnLockWriteFuture state")
                     .wakers
                     .push_back(context.waker().clone());
 
